import numpy as np
from heavylight import LightModel
from typing import Union, Callable
import pytest

class SimpleModel(LightModel):

    def __init__(self, initial_pols_if: np.ndarray, storage_function: Union[Callable, None] = None, mortality_rate = .01):
        super().__init__(storage_function=storage_function)
        self.initial_pols_if = initial_pols_if
        self.mortality_rate = mortality_rate

    def t(self, t):
        return t

    def num_pols_if(self, t):
        if t == 0:
            return self.initial_pols_if
        else:
            return self.num_pols_if(t - 1) - self.pols_death(t - 1) # causes exponential time complexity if uncached
        
    def pols_death(self, t):
        return self.num_pols_if(t) * self.mortality_rate

    def cashflow(self, t):
        return self.num_pols_if(t) * 100
    
    def v(self, t):
        """discount factor for time t --> time 0"""
        if t == 0:
            return 1
        else:
            return self.v(t - 1) / (1 + self.forward_rate(t))
    
    def forward_rate(self, t):
        return 0.04
    
    def pv_cashflow(self, t):
        """present value of the cashflow occuring at time t"""
        return self.cashflow(t) * self.v(t)
    
    def fib(self, t, mult_factor):
        if t <= 1:
            return t
        return mult_factor * self.fib(t - 1, mult_factor) + self.fib(t - 2, mult_factor)

def test_forward_rate():
    sm = SimpleModel(np.linspace(.1, 1, 10), np.sum)
    sm.forward_rate(0)
    assert sm.forward_rate.cache[((0,), frozenset())] == .04
    assert len(sm.StoredResults()) == 1
    assert len(sm.StoredResults()['forward_rate']) == 1
    assert sm.StoredResults()['forward_rate'][0] == .04

    sm.forward_rate(5)
    assert len(sm.forward_rate.cache) == 2


def test_timestep_functions():
    sm = SimpleModel(np.linspace(.1, 1, 10), np.sum)
    timestep_functions = sm.TimestepFunctions()
    expected_functions = ['t', 'num_pols_if', 'pols_death', 'cashflow', 'v', 'pv_cashflow', 'forward_rate']
    assert set(timestep_functions) == set(expected_functions) # no duplicates
    sm.RunModel(5)
    # The timestep functions are the ones displayed in StoredResults()
    assert set(sm.StoredResults().keys()) == set(expected_functions)

@pytest.mark.timeout(4)
def test_caching():
    sm = SimpleModel(np.linspace(.1, 1, 10), np.sum)
    assert len(sm.num_pols_if.cache) == 0
    sm.RunModel(200)
    assert len(sm.num_pols_if.cache) == 201
    assert type(sm.StoredResults()['num_pols_if'][200]) == np.float64
    # fib did not get called because it is not single parameter time function
    assert len(sm.fib.cache) == 0
    sm.fib(200, 1.1)
    # but it is still cached
    assert len(sm.fib.cache) == 201

def test_dataframe():
    sm = SimpleModel(np.linspace(.1, 1, 10), np.sum)
    sm.RunModel(5)
    df = sm.ToDataFrame()
    assert df.shape == (6, 7)
    assert df['t'].tolist() == [0, 1, 2, 3, 4, 5]

def test_reset_cache():
    sm = SimpleModel(np.linspace(.1, 1, 10), storage_function=np.sum, mortality_rate=.01)
    sm.RunModel(5)
    assert round(sm.StoredResults()['pols_death'][0], 10) == .055
    sm.mortality_rate = .02
    sm.RunModel(5)
    assert round(sm.StoredResults()['pols_death'][0], 10) == .055
    sm.ResetCache()
    sm.RunModel(5)
    assert round(sm.StoredResults()['pols_death'][0], 10) == .11
    




from inspect import getmembers, signature
from typing import Callable, List, Union
from types import MethodType
from heavylight.memory_optimized_cache import CacheGraph, _Cache
import pandas as pd


class LightModel:
    """Base class to subclass from for recursive actuarial models.

    Inheriting from this class causes functions starting with a lowercase letter to be cached.
    A cached function will only be called once for each unique set of arguments.
    Caching is necessary for recursive actuarial calculations to be efficient.

    This means that the function will only be called once for each unique set of arguments, important for recursive actuarial models.

    Parameters
    ----------
    storage_function: Callable[[int], Any], optional
        This function is applied to the results of cached methods that have a single parameter `t`.
        (e.g. `def pols_if(self, t):`).  The default is None, no values will be stored.
        `storage_function=np.sum` will cause the sum of the results to be stored.
        `storage_function=lambda x: x[0] if isinstance(x, np.ndarray) else x` will store the first element.
        `storage_function=lambda x: x` will store the raw results.

    Class level methods:
        RunModel(proj_len):
        if the model has not been auto-run at initialisation, run it for projection length.

    methods/variables to avoid:
    methods/variables starting with an underscore `_` are treated as internal.  You may break functionality if you create your own.
    """
    
    def __init__(self, storage_function: Union[Callable, None] = None):
        self.cache_graph = CacheGraph()
        self._single_param_timestep_funcs: List[_Cache]  = []
        self._storage_function = storage_function
        # happens after setting up attributes
        for method_name, method in getmembers(self):
            if not method_name[0].islower() or method_name.startswith("_") or not isinstance(method, MethodType):
                continue
            is_single_param_t = check_if_single_parameter_t(method)
            if is_single_param_t:
                cached_method = self.cache_graph(self._storage_function)(method)
                self._single_param_timestep_funcs.append(cached_method)
            else:
                cached_method = self.cache_graph()(method)
            setattr(self, method_name, cached_method)

    def RunModel(self, proj_len: int):
        """
        Run the model for a projection length.
        All single parameter timestep functions will be run for each timestep.
        """
        for t in range(proj_len+1):
            for func in self._single_param_timestep_funcs:
                func(t)

    def ResetCache(self):
        """Reset the cache, useful if you want to run the model again with different parameters."""
        self.cache_graph.reset()

    def OptimizeMemoryAndReset(self):
        """
        Calling this clears the cache, and causes the cache to be optimized.
        Optimizing the cache means that only needed results are stored, and intermediate results are cleared.
        """
        self.cache_graph.optimize_and_reset()

    def TimestepFunctions(self):
        """List the cached functions that have a single parameter `t`."""
        return [cache._func.__name__ for cache in self._single_param_timestep_funcs]
    
    def StoredResults(self):
        """
        Return a dictionary of all stored results. 
        The dictionary is of the form:
        {function_name: {timestep: stored_result}}
        """
        return self.cache_graph.stored_results
    
    def ToDataFrame(self):
        """Return a pandas dataframe of all single parameter timestep results."""
        df = pd.DataFrame(self.StoredResults())
        # if t is in the dataframe, move it to first position
        if "t" in df.columns:
            df.insert(0, "t", df.pop("t"))
        return df


def check_if_single_parameter_t(func: Callable):
    sig = signature(func)
    return 't' in sig.parameters and len(sig.parameters) == 1


import types
from typing import List
import warnings
from inspect import signature
import pandas as pd

class Table:
    """A Table has one or more keys, and a single column of values"""

    def __init__(self, series):
        """Initialise a table with a series - for multiple keys use a multikey index"""
        self.series = series

    def __getitem__(self, key):
        return self.series.at[key]
    
    def __call__(self, key):
        return self.series.at[key]
    
    def values(self, key):
        return self.series.loc[key].values

    @staticmethod
    def from_csv(filename, sep=",", type_identifier = "|"):
        """Read in a table from a csv file, type encoding is via the header:
        
        <keyname1>|<type1>,<keyname2>|<type2>....<value>|<typeN>
        where type is one of "str", "int", "float"
        """
        column_types = {"str":str, "int": int, "float": float}
        with open(filename, "r") as csv_file:
            header = next(csv_file).strip("\n").split(sep)
        tid = type_identifier
        header_mapper_str = {item:item.split(tid)[1] for item in header}
        header_mapper_types = {col:column_types[val] for col, val in header_mapper_str.items()}
        df = pd.read_csv(filename, sep=sep, dtype=header_mapper_types)
        # strip `tid` from column names
        df.columns = [col.split("|")[0] for col in df.columns]
        df.set_index(list(df.columns[:-1]), inplace=True)
        series = df[df.columns[0]]
        return Table(series=series)


class _Cache:
    """Cache provides controllable memoization for model methods"""

    def __init__(self, func: types.MethodType, param_len: int):
        self.func = func
        self.param_len = param_len
        self.has_one_param = self.param_len == 1
        self.values = dict()
        self.__name__ = "Cache: " + func.__name__

    def __call__(self, *arg:list):
        if arg in self.values:
            return self.values[arg]
        else:
            result = self.func(*arg)
            self.values[arg] = result
            return result

    def __repr__(self) -> str:
        return f"<Cache Function: {self.func.__name__} Size: {len(self.values)}>"

class Model:
    def __init__(self, *, do_run: bool = False, verbose=False, proj_len: int|types.NoneType = None, **kwargs,):
        """Base Class to subclass for user models.

        All variables/methods in user models should be lower case, using underscore as spaces.

        Class level methods:
          RunModel(proj_len):
            if the model has not been auto-run at initialisation, run it for projection length.

        Special user methods:
          BeforeRun(self):
            If this is specified in the user model it called before the projection starts, e.g. to set up some specific variables

          AfterRun(self):
            user method, called after Run is completed, e.g. can use to calculate NPVs of variables

        methods/variables to avoid:
        methods/variables starting with an underscore `_` are treated as internal.  You may break functionality if you create your own.

        """

        self._cached = False
        self._is_run = False
        if verbose:
            print("== Run Parameters ==")
            print("    do_run:", do_run)
            print("    proj_len:", proj_len)
            print()
        
        if do_run:
            if not isinstance(proj_len, int):
                raise ValueError("proj_len must be an integer")
            elif proj_len <= 0:
                raise ValueError("proj_len must have value greater than 0")
            self.proj_len = proj_len
        else:
            if verbose: print("== Not Running - call Run() manually ==")
       
        if verbose: print("== Storing Arguments ==")
        for k, v in kwargs.items():
            if k in dir(self):
                warnings.warn("Warning: Duplicate Item: "+str(k))
            setattr(self, k, v)
            if verbose:
                print("    Updated: ", k, " : ", v)

        # cacheify
        if verbose: print("== Caching Functions ==")
        self._cache_funcs(verbose)

        if do_run and proj_len > 0:
            self.RunModel(proj_len, verbose)
        if verbose: print("== Run complete == ")


    def RunModel(self, proj_len, verbose=False):
        if self._is_run:
            raise ValueError("Run has already been completed.")

        if verbose: print(f"== Running Projection | length: {proj_len} ==")
        
        if hasattr(self, "BeforeRun"):
            if verbose: print("    Calling BeforeRun")
            self.BeforeRun()

        if not self._cached:
            raise ValueError("Functions have not been cached")  # NB: this shouldn't occur as now caching in instance
        for t in range(proj_len):
            for var in self._funcs.keys():
                func = getattr(self, var)
                if func.has_one_param:   # skip functions with more than one parameter
                    func(t)   #call each function in turn, starting from t==0
        self._is_run = True
        if hasattr(self, "AfterRun"):
            if verbose: print("    Calling AfterRun")
            return self.AfterRun()
    
    def _cache_funcs(self, verbose : bool=False):
        if self._cached:
            raise ValueError("Cache has already been set-up, please create a new instance")

        self._funcs = {}

        for method_name in dir(self):
            method = getattr(self, method_name)

            if method_name[0] != "_" and method_name[0].islower() and isinstance(method, types.MethodType):
                param_count = len(signature(method).parameters) # count the parameters in the function.
                cached_method = _Cache(method, param_count)
                setattr(self, method_name, cached_method)
                self._funcs[method_name] = cached_method
                if verbose: print(f"    Cached: {method_name}")

        
        self._cached = True
        
    def ToDataFrame(self):
        """return a pandas dataframe of all single parameter columns"""
        df = pd.DataFrame()
        for func in self._funcs:
            if self._funcs[func].has_one_param:
                df[func] = pd.Series(self._funcs[func].values)
        return df